import socket
from time import sleep
import datetime
import logging
from typing import Optional

from .jvccommands import *

_LOGGER = logging.getLogger(__name__)


class JVCProjector:
    """JVC Projector Control"""

    def __init__(
        self,
        host: str,
        password: Optional[str] = None,
        port: Optional[int] = None,
        delay_ms: Optional[int] = None,
        connect_timeout: Optional[int] = None,
        max_retries: Optional[int] = None,
    ):
        self.host = host
        self.port = port if port else 20554
        self.connect_timeout = connect_timeout if connect_timeout else 10
        self.delay = (
            datetime.timedelta(microseconds=(delay_ms * 1000))
            if delay_ms
            else datetime.timedelta(microseconds=(600000))
        )
        self.last_command_time = datetime.datetime.now() - datetime.timedelta(
            seconds=10
        )
        self.password = password if password else ""
        self.max_retries = max_retries if max_retries else 10

        _LOGGER.debug(
            f"initialising JVCProjector with host={self.host}, password={self.password}, port={self.port}, delay={self.delay.total_seconds()*1000}, connect_timeout={self.connect_timeout}, max_retries={self.max_retries}"
        )

        self.JVC_REQ = b"PJREQ"
        if self.password != None and len(self.password):
            if len(self.password) == 10:
                self.JVC_REQ = b"PJREQ_" + bytes(self.password, "ascii")
            elif len(self.password) == 9:
                self.JVC_REQ = b"PJREQ_" + bytes(self.password, "ascii") + b"\x00"
            elif len(self.password) == 8:
                self.JVC_REQ = b"PJREQ_" + bytes(self.password, "ascii") + b"\x00\x00"
            else:
                raise JVCConfigError(
                    "Specified network password invalid (too long/short). Please check your configuration."
                )
        # run validate_connection just but do nothing if it fails, as communication attempts may succeed later.
        self.validate_connection()

    def __throttle(self, last_time: datetime.datetime) -> None:
        if self.delay == 0:
            return

        delta = datetime.datetime.now() - last_time

        if self.delay > delta:
            sleep((self.delay - delta).total_seconds() * 1.1)

    def __connect(self, retry: int = 0) -> socket.socket:
        jvc_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        jvc_sock.settimeout(self.connect_timeout)
        try:
            jvc_sock.connect((self.host, self.port))
            return jvc_sock
        except socket.timeout as e:
            jvc_sock.close()
            raise JVCCannotConnectError(
                "Timed out when trying to connect to projector"
            ) from e
        except OSError as e:
            jvc_sock.close()
            if retry < self.max_retries:
                _LOGGER.debug(
                    f"Received error: {repr(e)} when trying to connect, retrying (we're on retry number {retry+1} of {self.max_retries})"
                )
                self.__throttle(datetime.datetime.now())
                jvc_sock = self.__connect(retry + 1)
                _LOGGER.debug(f"Connection successful")
                return jvc_sock
            raise JVCCannotConnectError(
                "Could not establish connection to projector"
            ) from e

    def __handshake(self) -> socket.socket:

        JVC_GRT = b"PJ_OK"
        JVC_REQ = self.JVC_REQ
        JVC_ACK = b"PJACK"

        jvc_sock = self.__connect()

        _LOGGER.debug(f"Attempting handshake")

        try:
            message = jvc_sock.recv(len(JVC_GRT))
            if message != JVC_GRT:
                jvc_sock.close()
                raise JVCHandshakeError(
                    f"Projector did not reply with PJ_OK (got `{message}` instead)"
                )

        except socket.timeout as e:
            jvc_sock.close()
            raise JVCHandshakeError("Timeout when waiting to receive PJ_OK") from e

        # try sending PJREQ, if there's an error, raise exception
        try:
            jvc_sock.sendall(JVC_REQ)
        except OSError as e:
            jvc_sock.close()
            raise JVCHandshakeError("Socket exception when sending PJREQ.") from e

        # see if we receive PJACK, if not, raise exception
        try:
            message = jvc_sock.recv(len(JVC_ACK))
            if message != JVC_ACK:
                jvc_sock.close()
                raise JVCHandshakeError(
                    f"Projector did not reply with PJACK (received `{message}` instead)"
                )
        except socket.timeout as e:
            jvc_sock.close()
            raise JVCHandshakeError("Timeout when waiting to receive PJACK") from e

        _LOGGER.debug(f"Handshake successful")
        return jvc_sock

    def _send_command(self, command: Command, value: str = "") -> Optional[str]:
        """Call Commands.read() if not value, else Commands.write(value)"""

        self.__throttle(self.last_command_time)

        jvc_sock: socket.socket = self.__handshake()

        result: Optional[str] = None

        if value:
            _LOGGER.debug(
                f"writing property: {jfrmt.highlight(value)} to command group: {jfrmt.highlight(command.name)}"
            )
            command.write(jvc_sock, value)
        elif command.write_only:
            _LOGGER.debug(
                f"sending write_only operation: {jfrmt.highlight(command.name)}"
            )
            command.write(jvc_sock)
        else:
            _LOGGER.debug(
                f"reading from command group: {jfrmt.highlight(command.name)}"
            )
            result = command.read(jvc_sock)

        self.last_command_time = datetime.datetime.now()
        _LOGGER.debug(f"command sent successfully")

        if result is not None:
            _LOGGER.debug(
                f"the state of command group: {jfrmt.highlight(command.name)} is: {jfrmt.highlight(result)}"
            )
            return result

    def validate_connection(self) -> bool:
        try:
            _LOGGER.debug(
                f"Sending nullcmd to {jfrmt.highlight(f'{self.host}:{self.port}')} to verify connection"
            )
            self._send_command(Commands.nullcmd)
            return True
        except JVCCannotConnectError as e:
            _LOGGER.warning(f"Couldn't verify connection to projector at the specified address: {self.host}:{self.port}.")
            return False


    def power_on(self) -> None:
        self._send_command(Commands.power, "on")

    def power_off(self) -> None:
        self._send_command(Commands.power, "off")

    def command(self, command_string: str) -> Optional[str]:
        ps = self.power_state()
        if (command_string not in ["power-on", "power"]) and ps != "lamp_on":
            raise JVCPoweredOffError(
                f"Can't execute command: `{command_string}` unless the projector is in state `lamp_on`. "
                f"Current power state is: `{ps}`"
            )

        commandl: list[str] = command_string.split("-")

        if not hasattr(Commands, commandl[0]):
            raise JVCCommandNotFoundError(
                f"The requested command: `{command_string}` is not in the list of recognised commands"
            )
        else:
            if len(commandl) > 1:
                self._send_command(getattr(Commands, commandl[0]), commandl[1])
            else:
                return self._send_command(getattr(Commands, commandl[0]))

    def power_state(self) -> Optional[str]:
        return self._send_command(Commands.power)

    def is_on(self) -> bool:
        on = ["lamp_on", "reserved"]
        return self.power_state() in on



"""Commands for the jvc_projector library."""

import socket
import logging
from dataclasses import dataclass, field

_LOGGER = logging.getLogger(__name__)

# Headers
OPR = b"!\x89\x01"  # operation (set)
REF = b"?\x89\x01"  # reference (get)
RES = b"@\x89\x01"  # response
ACKH = b"\x06\x89\x01"  # projector ack
COM_ACK_LENGTH = 6  # length of ACKs sent by the projector


# use a dataclass so that we have a nice repr
@dataclass(init=False)
class Command:
    """Base class for defining a set of read/write commands.
    Args:
        cmd (bytes): the base command bytes, for examplee b"PW" for power, b"PMPM" for picture mode.
        *args (dict[str, bytes]): dictionaries of str:bytes pairs defining the read and write values.
            If two dicts are provided as args, then the first is used for the write commands and
            the second is used for read values. If only a single dict is provided,
            we use it for both the write commands and the read values.
        write_only(bool, optional): If true, this command group does not read from the projector
            Defaults to False.
        read_only(bool, optional): If True, this command group cannot write to the projector
            Defaults to False.
        verify_write (bool, optional): Whether we should wait for an ACK once we send a write command.
            Defaults to True.

    Examples:
        See jvccommands.Commands, jvcprojector.JVCProjector
    """

    cmd: bytes
    name: str
    readwritevals: tuple[dict[str, bytes]] = field(repr=False)
    write_only: bool
    read_only: bool
    verify_write: bool

    write_vals: dict[str, bytes]
    read_vals: dict[str, bytes]

    def __init__(
        self,
        cmd: bytes,
        *readwritevals: dict[str, bytes],
        write_only: bool = False,
        read_only: bool = False,
        verify_write: bool = True,
    ):
        self.cmd = cmd
        self.verify_write = verify_write

        self.ack = ACKH + self.cmd[0:2] + b"\n"

        try:
            assert len(readwritevals) <= 2
        except AssertionError:
            raise AssertionError(
                "(set_vals, get_vals) AND setget_vals cannot be defined at the same time."
            )

        self.write_only = write_only
        self.read_only = read_only
        if len(readwritevals) == 1:
            if write_only:
                self.write_vals = readwritevals[0]
                self.read_vals = {}
            elif read_only:
                self.write_vals = {}
                self.read_vals = readwritevals[0]
            else:
                self.write_vals = readwritevals[0]
                self.read_vals = readwritevals[0]

        elif len(readwritevals) == 2:
            self.write_vals = readwritevals[0]
            self.read_vals = readwritevals[1]
        else:
            self.write_vals = {}
            self.read_vals = {}

        self.write_valsinv = {
            self.write_vals[key]: key for key in self.write_vals.keys()
        }
        self.read_valsinv = {self.read_vals[key]: key for key in self.read_vals.keys()}

    # TODO: this only works if the commands are defined within a class.
    # A globally defined Command won't have the`name` property.
    def __set_name__(self, owner, name: str):
        self.name = name

    def __send(self, sock: socket.socket, command: bytes) -> None:
        try:
            sock.sendall(command)
        except OSError as e:
            sock.close()
            raise JVCCommunicationError(
                f"Socket exception of `{self.name}` command when sending bytes: `{command}`."
            ) from e

    def __verify_ack(self, sock: socket.socket, command: bytes) -> None:
        try:
            ACK = sock.recv(COM_ACK_LENGTH)

            # check if the ACK is valid (compare to user provided ack if available)
            if not ACK.startswith(ACKH):
                sock.close()
                raise JVCCommunicationError(
                    f"Malformed ACK response from the projector when sending command: `{self.name}` with bytes: `{command}`. "
                    f"Received ACK: `{ACK}` does not have the correct header"
                )
            elif not ACK == self.ack:
                sock.close()
                raise JVCCommunicationError(
                    f"Malformed ACK response from the projector when sending command: `{self.name}` with bytes: `{command}`. "
                    f"Expected `{self.ack}`, received `{ACK}`"
                )

        except socket.timeout as e:
            sock.close()
            raise JVCCommunicationError(
                f"Timeout when waiting for the specified ACK: `{self.ack}` for command: `{self.name}` with bytes: `{command}`"
            ) from e

        except OSError as e:
            sock.close()
            raise JVCCommunicationError(
                f"Socket exception when waiting for the specified ACK: `{self.ack}` for command: `{self.name}` with bytes: `{command}`"
            ) from e

    def write(self, sock: socket.socket, value: str = "") -> None:
        if self.read_only:
            sock.close()
            raise JVCCommandNotFoundError(
                f"The command group `{self.name}` does not implement any writable properties"
            )
        try:
            command = OPR + self.cmd + self.write_vals[value] + b"\n"
        except KeyError as e:
            if not value and not self.write_vals:
                command = OPR + self.cmd + b"\n"
            elif not value and self.write_only:
                sock.close()
                raise JVCCommandNotFoundError(
                        f"Write only command group: `{self.name}` has to be called with a corresponding key to complete the write operation. "
                        f"Must be one of: {list(self.write_vals.keys())}."
                ) from e
            else:
                sock.close()
                raise JVCCommandNotFoundError(
                    f"The command: `{self.name}` does not contain operation: `{value}`. "
                    f"Must be one of: {list(self.write_vals.keys())}."
                ) from e

        if not self.verify_write:
            _LOGGER.debug(
                f"ACK verification disabled for the command: `{jfrmt.highlight(self.name)} `. Error handling will be less robust"
            )

        self.__send(sock, command)

        # no need to wait for ACK or message as this is not a reference command without ACK specified
        if not self.verify_write:
            sock.close()
            return

        self.__verify_ack(sock, command)

        sock.close()

    def read(self, sock: socket.socket) -> str:
        if self.write_only:
            sock.close()
            raise JVCCommandNotFoundError(
                f"The command group: `{self.name}` does not implement any readable properties"
            )

        command = REF + self.cmd + b"\n"

        self.__send(sock, command)

        self.__verify_ack(sock, command)

        try:
            resp = sock.recv(1024)
        except socket.timeout as e:
            sock.close()
            raise JVCCommunicationError(
                f"Timeout when waiting for response for read command: `{self.name}`"
            ) from e
        except OSError as e:
            sock.close()
            raise JVCCommunicationError(
                f"Socket exception when waiting for response for read command: `{self.name}`"
            ) from e

        sock.close()
        try:
            assert resp.startswith(RES + self.cmd[0:2])
        except AssertionError as e:
            raise JVCCommunicationError(
                f"Malformed response header for read command: `{self.name}`"
            ) from e

        resp = resp[len(RES) + 2 : -1]

        # read_vals not defined but we still received a response
        resp_ascii = resp.decode("ascii")
        if not self.read_vals:
            return resp_ascii

        # decode the response if it's known, otherwise just return the
        # raw ascii code and log a warning.
        try:
            return self.read_valsinv[resp]
        except KeyError as e:
            _LOGGER.warning(
                    f"Could not decode response: `{resp_ascii}` for command: `{jfrmt.highlight(self.name)}`. "
                    f"It is not in the list of known responses, returning the raw ascii value instead."
            )
            return resp_ascii


class Commands:
    """A container for Commands"""

    # power commands
    power = Command(
        b"PW",
        {"on": b"1", "off": b"0"},
        {
            "standby": b"0",
            "lamp_on": b"1",
            "cooling": b"2",
            "reserved": b"3",
            "emergency": b"4",
        },
    )

    # lens memory commands
    memory = Command(
        b"INML",
        {"1": b"0", "2": b"1", "3": b"2", "4": b"3", "5": b"4", "6": b"5", "7": b"6", "8": b"7", "9": b"8", "10": b"9"},
    )

    # input commands, input is technically a keyword, but should be okay...
    input = Command(b"IP", {"hdmi1": b"6", "hdmi2": b"7"})

    # picture mode commands
    picture_mode = Command(
        b"PMPM",
        {
            "film": b"00",
            "cinema": b"01",
            "natural": b"03",
            "hdr10": b"04",
            "thx": b"06", # not present in NZ series
            "frame_adapt_hdr": b"0B", # new in NZ series
            "user1": b"0C",
            "user2": b"0D",
            "user3": b"0E",
            "user4": b"0F",
            "user5": b"10",
            "user6": b"11",
            "hlg": b"14",
            "hdr10p": b"15", # new in NZ series
            "pana_pq": b"16", # new in NZ series
        },
    )

    # low latency enable/disable
    low_latency = Command(b"PMLL", {"on": b"1", "off": b"0"})

    # mask commands
    mask = Command(
        b"ISMA",
        {"off": b"2", "custom1": b"0", "custom2": b"1", "custom3": b"3"},
    )

    # lamp commands
    lamp = Command(
        b"PMLP",
        {"high": b"1", "low": b"0", "mid": b"2"},
    )

    # menu controls
    menu = Command(
        b"RC73",
        {
            "menu": b"2E",
            "down": b"02",
            "left": b"36",
            "right": b"34",
            "up": b"01",
            "ok": b"2F",
            "back": b"03",
        },
        write_only=True,
    )

    # Intelligent Lens Aperture commands
    aperture = Command(
        b"PMDI",
        {"off": b"0", "auto1": b"1", "auto2": b"2"},
    )

    # Anamorphic commands
    anamorphic = Command(
        b"INVS",
        {"off": b"0", "a": b"1", "b": b"2", "c": b"3", "d": b"4"},
    )

    # active signal
    signal = Command(
        b"SC",
        {"no_signal": b"0", "active_signal": b"1"},
        read_only=True,
    )

    # MAC address, model, null command
    macaddr = Command(
        b"LSMA",
        read_only=True,
    )

    modelinfo = Command(
        b"MD",
        read_only=True,
    )
    nullcmd = Command(
        b"\x00\x00",
        write_only=True,
    )


class JVCConfigError(Exception):
    """Exception when the user supplied config is wrong"""

    pass


class JVCCannotConnectError(Exception):
    """Exception when we can't connect to the projector"""

    pass


class JVCHandshakeError(Exception):
    """Exception when there was a problem with the 3 step handshake"""

    pass


class JVCCommunicationError(Exception):
    """Exception when there was a communication issue"""

    pass


class JVCCommandNotFoundError(Exception):
    """Exception when the requested command doesn't exist"""

    pass


class JVCPoweredOffError(Exception):
    """Exception when projector is powered off and can't accept some commands."""

    pass

class jfrmt:
    @staticmethod
    def highlight(value: str) -> str:
        return "{:s}".format("\u035F".join(value))



import socket
from enum import Enum
from time import sleep
import datetime


class Commands(Enum):
    # power commands
    power_on = b"\x21\x89\x01\x50\x57\x31\x0A"
    power_off = b"\x21\x89\x01\x50\x57\x30\x0A"

    # lens memory commands
    memory1 = b"\x21\x89\x01\x49\x4E\x4D\x4C\x30\x0A"
    memory2 = b"\x21\x89\x01\x49\x4E\x4D\x4C\x31\x0A"
    memory3 = b"\x21\x89\x01\x49\x4E\x4D\x4C\x32\x0A"
    memory4 = b"\x21\x89\x01\x49\x4E\x4D\x4C\x33\x0A"
    memory5 = b"\x21\x89\x01\x49\x4E\x4D\x4C\x34\x0A"

    # input commands
    hdmi1 = b"\x21\x89\x01\x49\x50\x36\x0A"
    hdmi2 = b"\x21\x89\x01\x49\x50\x37\x0A"

    # power status query commands
    power_status   = b"\x3F\x89\x01\x50\x57\x0A"
    current_output = b"\x3F\x89\x01\x49\x50\x0A"
    
    # picture mode commands
    pm_cinema = b"\x21\x89\x01\x50\x4D\x50\x4D\x30\x31\x0A"
    pm_hdr = b"\x21\x89\x01\x50\x4D\x50\x4D\x30\x34\x0A"
    pm_natural = b"\x21\x89\x01\x50\x4D\x50\x4D\x30\x33\x0A"
    pm_film = b"\x21\x89\x01\x50\x4D\x50\x4D\x30\x30\x0A"
    pm_THX = b"\x21\x89\x01\x50\x4D\x50\x4D\x30\x36\x0A"
    pm_user1 = b"\x21\x89\x01\x50\x4D\x50\x4D\x30\x43\x0A"
    pm_user2 = b"\x21\x89\x01\x50\x4D\x50\x4D\x30\x44\x0A"
    pm_user3 = b"\x21\x89\x01\x50\x4D\x50\x4D\x30\x45\x0A"
    pm_user4 = b"\x21\x89\x01\x50\x4D\x50\x4D\x30\x46\x0A"
    pm_user5 = b"\x21\x89\x01\x50\x4D\x50\x4D\x31\x30\x0A"
    pm_user6 = b"\x21\x89\x01\x50\x4D\x50\x4D\x31\x31\x0A"
    pm_hlg = b"\x21\x89\x01\x50\x4D\x50\x4D\x31\x34\x0A"

    # low latency enable/disable
    pm_low_latency_enable = b"\x21\x89\x01\x50\x4D\x4C\x4C\x31\x0A"
    pm_low_latency_disable = b"\x21\x89\x01\x50\x4D\x4C\x4C\x30\x0A"

    # mask commands
    mask_off = b"\x21\x89\x01\x49\x53\x4D\x41\x32\x0A"
    mask_custom1 = b"\x21\x89\x01\x49\x53\x4D\x41\x30\x0A"
    mask_custom2 = b"\x21\x89\x01\x49\x53\x4D\x41\x31\x0A"
    mask_custom3 = b"\x21\x89\x01\x49\x53\x4D\x41\x33\x0A"

    #lamp commands
    lamp_high = b"\x21\x89\x01\x50\x4D\x4C\x50\x31\x0A"
    lamp_low = b"\x21\x89\x01\x50\x4D\x4C\x50\x30\x0A"

    #menu controls
    menu = b"\x21\x89\x01\x52\x43\x37\x33\x32\x45\x0A"
    menu_down = b"\x21\x89\x01\x52\x43\x37\x33\x30\x32\x0A"
    menu_left = b"\x21\x89\x01\x52\x43\x37\x33\x33\x36\x0A"
    menu_right = b"\x21\x89\x01\x52\x43\x37\x33\x33\x34\x0A"
    menu_up = b"\x21\x89\x01\x52\x43\x37\x33\x30\x31\x0A"
    menu_ok = b"\x21\x89\x01\x52\x43\x37\x33\x32\x46\x0A"
    menu_back = b"\x21\x89\x01\x52\x43\x37\x33\x30\x33\x0A"

    #Lens Aperture commands
    aperture_off = b"\x21\x89\x01\x50\x4D\x44\x49\x30\x0a"
    aperture_auto1 = b"\x21\x89\x01\x50\x4D\x44\x49\x31\x0a"
    aperture_auto2 = b"\x21\x89\x01\x50\x4D\x44\x49\x32\x0a"

    #Anamorphic commands
    anamorphic_off = b"\x21\x89\x01\x49\x4E\x56\x53\x30\x0A"
    anamorphic_a = b"\x21\x89\x01\x49\x4E\x56\x53\x31\x0A"
    anamorphic_b = b"\x21\x89\x01\x49\x4E\x56\x53\x32\x0A"
    anamorphic_c = b"\x21\x89\x01\x49\x4E\x56\x53\x33\x0A"



class PowerStates(Enum):
    standby   = b"\x40\x89\x01\x50\x57\x30\x0A"
    cooling   = b"\x40\x89\x01\x50\x57\x32\x0A"
    emergency = b"\x40\x89\x01\x50\x57\x34\x0A"

    # on some projectors like the DLA-X5900, the status
    # is returned as the "reserved" on below when the
    # projector lamp is warming up and "lamp_on" when
    # the lamp is on
    lamp_on  = b"\x40\x89\x01\x50\x57\x31\x0A"
    reserved = b"\x40\x89\x01\x50\x57\x33\x0A"


class ACKs(Enum):
    power_ack = b"\x06\x89\x01\x50\x57\x0A"
    input_ack = b"\x06\x89\x01\x49\x50\x0A"


class JVCProjector:
    """JVC Projector Control"""

    def __init__(self, host, password = None, port=20554, delay_ms=600, connect_timeout=60):
        self.host = host
        self.port = port
        self.connect_timeout = connect_timeout
        self.delay = datetime.timedelta(microseconds=(delay_ms * 1000))
        self.last_command_time = datetime.datetime.now() - datetime.timedelta(seconds=10)
        self.password = password

    def throttle(self):
        if self.delay == 0:
            return

        delta = datetime.datetime.now() - self.last_command_time

        if self.delay > delta:
            sleep((self.delay - delta).total_seconds())

        return

    def _send_command(self, operation, ack=None):
        JVC_GREETING = b'PJ_OK'
        if self.password != None:
            if len(self.password) == 10:
                JVC_REQ = b'PJREQ_' + bytes(self.password, 'ascii')
            elif len(self.password) == 9:
                JVC_REQ = b'PJREQ_' + bytes(self.password, 'ascii') + b'\x00'
            elif len(self.password) == 8:
                JVC_REQ = b'PJREQ_' + bytes(self.password, 'ascii') + b'\x00\x00'
            else:
                raise Exception("Specified network password invalid (too long/short)")
        else:
            JVC_REQ = b'PJREQ'
        JVC_ACK = b'PJACK'
        result = False

        self.throttle()

        jvc_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        jvc_sock.settimeout(self.connect_timeout)
        jvc_sock.connect((self.host, self.port)) # connect to projector

        # 3 step handshake:
        # Projector sends PJ_OK, client sends PJREQ, projector replies with PJACK
        # first, after connecting, see if we receive PJ_OK. If not, raise exception
        if jvc_sock.recv(len(JVC_GREETING)) != JVC_GREETING:
            raise Exception("Projector did not reply with correct PJ_OK greeting")

        # try sending PJREQ, if there's an error, raise exception
        try:
            jvc_sock.sendall(JVC_REQ)
        except socket.error as e:
            raise Exception("Socket exception when sending PJREQ")

        # see if we receive PJACK, if not, raise exception
        if jvc_sock.recv(len(JVC_ACK)) != JVC_ACK:
            raise Exception("Socket exception on PJACK")

        # 3 step connection is verified, send the command
        jvc_sock.sendall(operation)

        # if we send a command that returns info, the projector will send
        # an ack, followed by the message. Check to see if the ack sent by
        # projector is correct, then return the message.
        if ack:
            ACK = jvc_sock.recv(len(ack))

            if ACK == ack:
                message = jvc_sock.recv(1024)
                result = message

        jvc_sock.close()

        self.last_command_time = datetime.datetime.now()

        return result

    def power_on(self):
        self._send_command(Commands.power_on.value)

    def power_off(self):
        self._send_command(Commands.power_off.value)

    def command(self, command_string):
        if not hasattr(Commands, command_string):
            return False
        else:
            self._send_command(Commands[command_string].value)
            return True

    def power_state(self):
        message = self._send_command(Commands.power_status.value, ack=ACKs.power_ack.value)
        return PowerStates(message).name

    def is_on(self):
        on = [PowerStates.lamp_on.name, PowerStates.reserved.name]
        return self.power_state() in on

